/**
 * \author Thomas Buechler
 * \brief standard heap sort with heapify, heapsort and swap method
 */

#ifndef EDSM_HEAPSORT_DEF
#define EDSM_HEAPSORT_DEF


template<uint8_t t_width>
void swap( sdsl::int_vector<t_width> & a, unsigned int i, unsigned int j){
	unsigned int buf = a[i];
	a[i] = a[j];
	a[j] = buf;
}

/*


//###################################################################################################

template<uint8_t t_width>
void heapify ( sdsl::int_vector<t_width> & a, unsigned int l, unsigned int r ) {
	unsigned int k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if (k+1 > r) {
		if(a[k] > a[l])
			swap(a,k,l);
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		heapify(a,k,r);
	}
}

template<uint8_t t_width>
void heapsort(sdsl::int_vector<t_width> & a){
	if(a.size()<2) return;
	
	unsigned int l =( (a.size() +1 ) / 2);	
	unsigned int r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,l,r);
	}
	
	while(r > 0){
		swap(a,r,0);
		r--;
		heapify(a,0,r);
	}
}*/

//###################################################################################################

//###################################################################################################
/*

void swap( std::deque<ui> & a, unsigned int i, unsigned int j){
	unsigned int buf = a[i];
	a[i] = a[j];
	a[j] = buf;
}

void heapify ( std::deque<ui> & a, unsigned int l, unsigned int r ) {
	unsigned int k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if (k+1 > r) {
		if(a[k] > a[l])
			swap(a,k,l);
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		heapify(a,k,r);
	}
}

void heapsort(std::deque<ui> & a){
	if(a.size()<2) return;
	
	unsigned int l =( (a.size() +1 ) / 2);	
	unsigned int r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,l,r);
	}
	
	while(r > 0){
		swap(a,r,0);
		r--;
		heapify(a,0,r);
	}
}*/

//###################################################################################################

template<class T>
void swap( std::vector<T> & a, uint64_t i, uint64_t j){
	T buf = a[i];
	a[i] = a[j];
	a[j] = buf;
}

template<class T>
void heapify ( std::vector<T> & a, uint64_t l, uint64_t r ) {
	uint64_t k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if (k+1 > r) {
		if(a[k] > a[l])
			swap(a,k,l);
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		heapify(a,k,r);
	}
}

template<class T>
void heapsort(std::vector<T> & a){
	if(a.size()<2) return;
	
	uint64_t l =( (a.size() +1 ) / 2);	
	uint64_t r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,l,r);
	}
	
	while(r > 0){
		swap(a,r,0);
		r--;
		heapify(a,0,r);
	}
}

//###################################################################################################
//USED  BY READ_PROCESSOR : find_hotspots
template<uint8_t t_width, uint8_t b_width>
void heapify ( sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, unsigned int l, unsigned int r ) {
	unsigned int k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if (k+1 > r) {
		if(a[k] > a[l]){
			swap(a,k,l);
			swap(b,k,l);
		}
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		swap(b,l,k);
		heapify(a,b,k,r);
	}
}

template<uint8_t t_width, uint8_t b_width>
void heapsort(sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b){
	assert(a.size() == b.size());
	if(a.size()<2) return;
	
	unsigned int l =( (a.size() +1 ) / 2);	
	unsigned int r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,b,l,r);
	}
	
	while(r > 0){
		swap(a,r,0);
		swap(b,r,0);
		r--;
		heapify(a,b,0,r);
	}
}

//###################################################################################################
//USED  BY READ_PROCESSOR : start_aligner() with inverse == true

const bool HEAPSORT_INVERSE = true;
template<uint8_t t_width, uint8_t b_width, uint8_t c_width>
void heapify ( sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, sdsl::int_vector<c_width> & c, unsigned int l, unsigned int r, bool inverse ) {
	unsigned int k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if(inverse){
		if (k+1 > r) {
			if(a[k] < a[l]){
				swap(a,k,l);
				swap(b,k,l);
				swap(c,k,l);
			}
			return;
		}
		
		if(a[k] > a[k+1]) 
			k++;
		
		if( a[l] > a[k] ){
			swap(a,l,k);
			swap(b,l,k);
			swap(c,l,k);
			heapify(a,b,c,k,r,inverse);
		}
		
		
		return;
	}
	
	if (k+1 > r) {
		if(a[k] > a[l]){
			swap(a,k,l);
			swap(b,k,l);
			swap(c,k,l);
		}
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		swap(b,l,k);
		swap(c,l,k);
		heapify(a,b,c,k,r,inverse);
	}
	
	
}

template<uint8_t t_width, uint8_t b_width, uint8_t c_width>
void heapsort(sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, sdsl::int_vector<c_width> & c, bool inverse=false){
	assert(a.size() == b.size() && a.size() == c.size());
	if(a.size()<2) return;
	
	unsigned int l =( (a.size() +1 ) / 2);	
	unsigned int r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,b,c,l,r,inverse);
	}
	
	while(r > 0){
		swap(a,r,0);
		swap(b,r,0);
		swap(c,r,0);
		r--;
		heapify(a,b,c,0,r,inverse);
	}
}

/*
//###################################################################################################
template<uint8_t t_width, uint8_t b_width, uint8_t c_width, uint8_t d_width>
void heapify ( sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, sdsl::int_vector<c_width> & c, sdsl::int_vector<d_width> & d, unsigned int l, unsigned int r ) {
	unsigned int k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if (k+1 > r) {
		if(a[k] > a[l]){
			swap(a,k,l);
			swap(b,k,l);
			swap(c,k,l);
			swap(d,k,l);
		}
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		swap(b,l,k);
		swap(c,l,k);
		swap(d,l,k);
		heapify(a,b,c,d,k,r);
	}
}

template<uint8_t t_width, uint8_t b_width, uint8_t c_width, uint8_t d_width>
void heapsort(sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, sdsl::int_vector<c_width> & c, sdsl::int_vector<d_width> & d){
	assert(a.size() == b.size() && a.size() == c.size() && a.size() == d.size());
	if(a.size()<2) return;
	
	unsigned int l =( (a.size() +1 ) / 2);	
	unsigned int r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,b,c,d,l,r);
	}
	
	while(r > 0){
		swap(a,r,0);
		swap(b,r,0);
		swap(c,r,0);
		swap(d,r,0);
		r--;
		heapify(a,b,c,d,0,r);
	}
}


//###################################################################################################

template<uint8_t t_width>
void unify_intervals( sdsl::int_vector<t_width> & l, sdsl::int_vector<t_width> & r){
	assert(l.size() == r.size());
	heapsort(l);
	heapsort(r);
			
	unsigned int li = l[0];
	unsigned int ri = r[0];
	unsigned int j = 0;
			
	for(unsigned int i = 1; i < l.size(); i++){
		if(l[i] > ri){
			//report interval
			l[j] = li;
			r[j] = ri;
			j++;
			li = l[i];
			ri = r[i];						
		} else { 
			//fuse intervals
			ri = r[i];						
		}					
	}
	//report last interval
	l[j] = li;
	r[j] = ri;
	j++;	
	l.resize(j);
	r.resize(j);
};


*/

//###################################################################################################

//READ_PROCESSOR start_aligner()

void swap( std::vector<std::string> & a, unsigned int i, unsigned int j){
	a[i].swap(a[j]);
}

template<uint8_t t_width, uint8_t b_width>
void heapify ( sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, std::vector<std::string> & b2, unsigned int l, unsigned int r ) {
	unsigned int k = 2*l + 1;
	
	if( k > r) 
		return;
	
	if (k+1 > r) {
		if(a[k] > a[l]){
			swap(a,k,l);
			swap(b,k,l);
			swap(b2,k,l);
		}
		return;
	}
	
	if(a[k] < a[k+1]) 
		k++;
	
	if( a[l] < a[k] ){
		swap(a,l,k);
		swap(b,l,k);
		swap(b2,l,k);
		heapify(a,b,b2,k,r);
	}
}



template<uint8_t t_width, uint8_t b_width>
void heapsort(sdsl::int_vector<t_width> & a, sdsl::int_vector<b_width> & b, std::vector<std::string> & c){
	if( a.size() != b.size() ||  a.size() != c.size() ) throw std::invalid_argument("heapsort:a, b and c must have the same size");
	if(a.size()<2) return;
	
	unsigned int l =( (a.size() +1 ) / 2);	
	unsigned int r = a.size()-1;
	
	while( l > 0){
		l--;
		heapify(a,b,c,l,r);
	}
	
	while(r > 0){
		swap(a,r,0);
		swap(b,r,0);
		swap(c,r,0);
		r--;
		heapify(a,b,c,0,r);
	}
}



#endif


