#ifndef EDSM_LIB_MIVB_DEF
#define EDSM_LIB_MIVB_DEF
//NEEDED BY READ PROCESSOR


/**
 * \brief int_vector_buffer that closes file after writing a block
 * 
 * \template integer with
 */
template<uint8_t t_width, class int_type>
struct my_int_vector_buffer{
	static_assert(t_width <= 64 , "int_vector: width of must be at most 64bits.");
	static_assert(t_width % 8 ==0, "int_vector: width of must be multiple of 8");
private:
	const static size_t	DEFAULT_BUF_SIZE = 1024*1024; //1MB
	bool 	i_open;
	bool 	all_in_buffer;
	char*	buffer;
	size_t	buf_size;
	size_t	next_byte;
	uint8_t	byte_p_int;
	
public:
	std::string  	f_name;
	//Constructor
	my_int_vector_buffer();
	my_int_vector_buffer(std::string f_name, size_t buf_size = DEFAULT_BUF_SIZE);
	//methods;
	void		push(int_type x);
	void		write_buf_to_file();
	void		open(std::string f_name, size_t buf_size = DEFAULT_BUF_SIZE);
	void		close();
	void		delete_all();
	bool		is_open();
	bool		is_all_in_buffer();
	//	
	bool 	get_buffer(sdsl::int_vector<t_width>                 & target);
	bool 	get_buffer(std::vector<int_type>                     & target);
	bool 	get_buffer(std::vector<std::pair<int_type,int_type>> & target );
	//
	static bool read_from_file(std::string f_name, sdsl::int_vector<t_width>                 & target , size_t buf_size = DEFAULT_BUF_SIZE);
	static bool read_from_file(std::string f_name, std::vector<int_type>                     & target , size_t buf_size = DEFAULT_BUF_SIZE);
	static bool read_from_file(std::string f_name, std::vector<std::pair<int_type,int_type>> & target , size_t buf_size = DEFAULT_BUF_SIZE);
};








template<uint8_t t_width, class int_type>
my_int_vector_buffer<t_width,int_type>::my_int_vector_buffer(){
	i_open  = false;
}

template<uint8_t t_width, class int_type>
my_int_vector_buffer<t_width,int_type>::my_int_vector_buffer(std::string f_name, size_t buf_size): f_name(f_name), buf_size(buf_size){	
	assert((buf_size % 8) == 0);
	buffer		=  new char[buf_size/8];
	next_byte 	= 0;
	byte_p_int	= t_width/8;
	i_open = true;
	edsm_io::file_remove_silent(f_name);
	all_in_buffer = true;
}

template<uint8_t t_width, class int_type>
void my_int_vector_buffer<t_width,int_type>::push(int_type x){
	assert(i_open);
	if(next_byte + byte_p_int > buf_size/8)
		write_buf_to_file();		
	
	for(int_type i = 1; i <= byte_p_int; i++){
		buffer[next_byte + byte_p_int - i] = (char) x;
		x >>= 8;
	}	
	
	next_byte	+= byte_p_int;
}

template<uint8_t t_width, class int_type>
my_int_vector_buffer<t_width,int_type>& operator <<(my_int_vector_buffer<t_width,int_type>& ivb, const int_type x){ 
	ivb.push(x);
	return ivb;
}

template<uint8_t t_width, class int_type>
void my_int_vector_buffer<t_width,int_type>::write_buf_to_file(){
	assert(i_open);	
	std::ofstream ofs(f_name, std::ios_base::out|std::ios_base::app|std::ios::binary);
	ofs.write (buffer,next_byte);
	next_byte = 0;
	ofs.close();
	all_in_buffer = false;
}

template<uint8_t t_width, class int_type>
/** \brief writes buffer to file and closes stream*/
void my_int_vector_buffer<t_width,int_type>::close(){
	assert(i_open);
	i_open = false;		
	write_buf_to_file();		
	delete[] buffer; //Free buffer
}
template<uint8_t t_width, class int_type>
/** \brief writes buffer to file and closes stream*/
void my_int_vector_buffer<t_width,int_type>::delete_all(){	
	assert(i_open);
	i_open = false;		
	delete[] buffer; //Free buffer
	edsm_io::file_remove_silent(f_name); //delete File
}

template<uint8_t t_width, class int_type>
void my_int_vector_buffer<t_width,int_type>::open(std::string f_name, size_t buf_size){
	assert((buf_size % 8) == 0);
	assert(!i_open);
	this->f_name 	= f_name;
	this->buf_size	= buf_size;
	buffer		= new char[buf_size/8];
	next_byte 	= 0;
	byte_p_int	= t_width/8;
	i_open 		= true;
	edsm_io::file_remove_silent(f_name);
	all_in_buffer = true;
}


template<uint8_t t_width, class int_type>
bool my_int_vector_buffer<t_width,int_type>::is_open(){
	return i_open;
}

template<uint8_t t_width, class int_type>
bool  my_int_vector_buffer<t_width,int_type>::read_from_file(std::string f_name, sdsl::int_vector<t_width> & target, size_t buf_size){
	static_assert	(t_width		<= 64,	"int_vector: width of must be at most 64bits.");
	static_assert	(t_width % 8	== 0,	"int_vector: width of must be multiple of 8");

	
	//check file size
	std::ifstream ifs	(f_name, std::ifstream::ate | std::ifstream::binary);
	size_t f_size 		= ifs.tellg();
	ifs.seekg(0);	
	
	assert		((buf_size * 8) % t_width	== 0);
	assert		((f_size   * 8) % t_width	== 0);
	
	char* buffer		= new char[buf_size];
	uint8_t byte_p_int	= t_width/8;
	int_type next_num 	= 0;
	
	target = sdsl::int_vector<t_width>(f_size/byte_p_int,0,t_width);
	
	while(!ifs.eof()){
		ifs.read (buffer,buf_size);			
		for( int_type i = 0; i + byte_p_int <= buf_size && next_num < target.size() ; i += byte_p_int ){
			
			assert(next_num < target.size());
			int_type num = 0;
			for(int_type j = 0; j < byte_p_int; j++){
				num <<= 8;
				num += (unsigned char) buffer[i+j];
			}
			target[next_num++] = num;
		}	
	}

	delete[] buffer;
	ifs.close();
	return 0; //MAYBE USE THIS FOR FAILURE CHECK
};

template<uint8_t t_width, class int_type>
bool  my_int_vector_buffer<t_width,int_type>::read_from_file(std::string f_name, std::vector<int_type> & target, size_t buf_size){
	static_assert	(t_width		<= 64,	"int_vector: width of must be at most 64bits.");
	static_assert	(t_width % 8	== 0,	"int_vector: width of must be multiple of 8");

	
	//check file size
	std::ifstream ifs	(f_name, std::ifstream::ate | std::ifstream::binary);
	size_t f_size 		= ifs.tellg();
	ifs.seekg(0);	
	
	assert		((buf_size * 8) % t_width	== 0);
	assert		((f_size   * 8) % t_width	== 0);
	
	char* buffer		= new char[buf_size];
	uint8_t byte_p_int	= t_width/8;
	int_type next_num 	= 0;
	
	target = std::vector<int_type>(f_size/byte_p_int);
	
	while(!ifs.eof()){
		ifs.read (buffer,buf_size);			
		for( int_type i = 0; i + byte_p_int <= buf_size && next_num < target.size() ; i += byte_p_int ){
			
			assert(next_num < target.size());
			int_type num = 0;
			for(int_type j = 0; j < byte_p_int; j++){
				num <<= 8;
				num += (unsigned char) buffer[i+j];
			}
			target[next_num++] = num;
		}	
	}

	delete[] buffer;
	ifs.close();
	return 0; //MAYBE USE THIS FOR FAILURE CHECK
};

template<uint8_t t_width, class int_type>
bool  my_int_vector_buffer<t_width,int_type>::read_from_file(std::string f_name, std::vector<std::pair<int_type,int_type>> & target, size_t buf_size){
	static_assert	(t_width		<= 64,	"int_vector: width of must be at most 64bits.");
	static_assert	(t_width % 8	== 0,	"int_vector: width of must be multiple of 8");

	
	//check file size
	std::ifstream ifs	(f_name, std::ifstream::ate | std::ifstream::binary);
	size_t f_size 		= ifs.tellg();
	ifs.seekg(0);	
	
	assert		((buf_size * 8) % t_width	== 0);
	assert		((f_size   * 8) % t_width	== 0);
	
	char* buffer		= new char[buf_size];
	uint8_t byte_p_int	= t_width/8;
	int_type next_pair 	= 0;
	
	target = std::vector<std::pair<int_type,int_type>>(f_size/byte_p_int/2);
	
	while(!ifs.eof()){
		ifs.read (buffer,buf_size);			
		for( int_type i = 0; (size_t) (i + 2*byte_p_int) <= buf_size && next_pair < target.size() ; i += byte_p_int ){
			
			assert(next_pair < target.size());
			
			int_type first = 0;
			for(int_type j = 0; j < byte_p_int; j++){
				first <<= 8;
				first += (unsigned char) buffer[i+j];
			}
			i += byte_p_int;
			int_type second = 0;
			for(int_type j = 0; j < byte_p_int; j++){
				second <<= 8;
				second += (unsigned char) buffer[i+j];
			}
			
			
			target[next_pair++] = std::make_pair(first,second);
		}	
	}

	delete[] buffer;
	ifs.close();
	return 0; //MAYBE USE THIS FOR FAILURE CHECK
};
template<uint8_t t_width, class int_type>
bool  my_int_vector_buffer<t_width,int_type>::get_buffer(sdsl::int_vector<t_width> & target){
	target = sdsl::int_vector<t_width>( next_byte/byte_p_int ,0,0);
	
	int_type next_num 	= 0;
	for(int_type i = 0; i < next_byte; i += byte_p_int){
		int_type num = 0;
		
		for(int_type j = 0; j < byte_p_int; j++){
			num <<= 8;
			num += (unsigned char) buffer[i+j];
		}
		target[next_num++] = num;		
	}
	return 0;
};

template<uint8_t t_width, class int_type>
bool  my_int_vector_buffer<t_width,int_type>::get_buffer(std::vector<int_type> & target){
	target = std::vector<int_type>(next_byte/byte_p_int);
	
	int_type next_num 	= 0;
	for(int_type i = 0; i < next_byte; i += byte_p_int){
		int_type num = 0;
		
		for(int_type j = 0; j < byte_p_int; j++){
			num <<= 8;
			num += (unsigned char) buffer[i+j];
		}
		target[next_num++] = num;		
	}
	return 0;
};

template<uint8_t t_width, class int_type>
bool  my_int_vector_buffer<t_width,int_type>::get_buffer(std::vector<std::pair<int_type,int_type>> & target){
	target = std::vector<std::pair<int_type,int_type>>((next_byte/byte_p_int)/2);
	
	int_type next_pair 	= 0;
	for(int_type i = 0; i < next_byte; i += byte_p_int){	
		
		int_type first  = 0;
		int_type second = 0;
		
		for(int_type j = 0; j < byte_p_int; j++){
			first <<= 8;
			first += (unsigned char) buffer[i+j];
		}
		i += byte_p_int;
		for(int_type j = 0; j < byte_p_int; j++){
			second <<= 8;
			second += (unsigned char) buffer[i+j];
		}		
		
		target[next_pair++] = std::make_pair(first,second);
	}
	return 0;
};

template<uint8_t t_width, class int_type>
bool my_int_vector_buffer<t_width,int_type>::is_all_in_buffer(){
	return all_in_buffer;
};

#endif





